/*
Danger from the Deep - Open source submarine simulation
Copyright (C) 2003-2006  Thorsten Jordan, Luis Barrancos and others.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// sound
// subsim (C)+(W) Markus Petermann and Thorsten Jordan. SEE LICENSE

#include <SDL.h>
#include <SDL_mixer.h>
#include <sstream>
#include <string>
using namespace std;
#include "system.h"
#include "sound.h"

#define FADE_TIME 2000

bool sound::use_sound = true;

sound::sound(const string& filename) : mixchunk(0)
{
	if (!use_sound) return;
	mixchunk = Mix_LoadWAV(filename.c_str());
	if (mixchunk == NULL) {
		ostringstream oss;
		oss << "Unable to load WAV file: " << Mix_GetError();
		sys().add_console(oss.str());
	}
}



sound::~sound()
{
	//fixme: maybe stop playing before that
	if (mixchunk != 0)
		Mix_FreeChunk(mixchunk);
}



void sound::play(float volume) const
{
	if (volume > 0.0 && use_sound && mixchunk != 0)
	{	
		Mix_VolumeChunk(mixchunk, int(128*volume+0.5f));
		if (!Mix_PlayChannel(-1, mixchunk, 0)) {
			ostringstream oss;
			oss << "Unable to play WAV file: " << Mix_GetError();
			sys().add_console(oss.str());
		}
	}
}

void sound::play(const sea_object* player, const sea_object* noise_source, const bool loop_forever)
{
	_play(player, noise_source, loop_forever, false);
}

void sound::play_fade(const sea_object* player, const sea_object* noise_source, const bool loop_forever)
{
	_play(player, noise_source, loop_forever, true);
}

void sound::_play(const sea_object* player, const sea_object* noise_source, const bool loop_forever,const  bool fade)
{
	if (use_sound && NULL != mixchunk)
	{	
		double distanceFromPlayer = player->get_pos().distance(noise_source->get_pos());
		// calculate max hearing distance (meters)
		double hearing_range = 0.0;
		double hearing_increment = 0.0;
		
		// audible range varies depending on whether player/noise source is submerged or not
		//
		// 255 == max dist value for Mix_SetPosition()
		if (player->get_depth() < 0)
		{
			// player submerged			
			if (noise_source->get_depth() < 0)
			{
				hearing_range = 20000.0;
				hearing_increment = 78.43; // hearing_range / 255;		
			}
			else
			{
				hearing_range = 30000.0;
				hearing_increment = 117.64; // hearing_range / 255;		
			}
		}
		else
		{
			// player on the surface
			if (noise_source->get_depth() < 0)
			{
				hearing_range = 5000.0;
				hearing_increment = 19.60; // hearing_range / 255;		
			}
			else
			{
				hearing_range = 10000.0;
				hearing_increment = 39.21; // hearing_range / 255;		
			}
		}
		
		// is sound within audible range
		if (distanceFromPlayer <= hearing_range)
		{						
			channel_num = -1;		
			angle bearingFromPlayer(noise_source->get_pos().xy() - player->get_pos().xy());					
			bearingFromPlayer = bearingFromPlayer + player->get_heading();				
			
			// dist for Mix_SetPosition() is calculated as a value between 0 and 255 with 0 being the loudest
			// (closest)			
			int dist = (int)(distanceFromPlayer / hearing_increment);
			
			if (true == fade)
				channel_num = Mix_FadeInChannel(-1, mixchunk, (loop_forever) ? -1 : 0, FADE_TIME);
			else
				channel_num = Mix_PlayChannel(-1, mixchunk, (loop_forever) ? -1 : 0);
																			
			if (channel_num)
			{		
				if (!Mix_SetPosition(channel_num, (short)bearingFromPlayer.value(), dist))
				{
					ostringstream oss;
					oss << "Mix_SetPosition() failed";
					sys().add_console(oss.str());
				}				
			}
			else
			{
				ostringstream oss;
				oss << "Unable to play WAV file: " << Mix_GetError();
				sys().add_console(oss.str());
			}
		}
	}
}

void sound::stop()
{
	Mix_HaltChannel(channel_num);
}

void sound::fade_out()
{
	Mix_FadeOutChannel(channel_num, FADE_TIME);
}

void sound::pause()
{
	Mix_Pause(channel_num);
}

void sound::resume()
{
	Mix_Resume(channel_num);
}

void sound::pause_all()
{
	Mix_Pause(-1);
}

void sound::resume_all()
{
	Mix_Resume(-1);
}

void sound::init()
{
	if (!use_sound) return;

	if (SDL_Init(SDL_INIT_AUDIO) != 0) {
		ostringstream oss;
		oss << "Unable to to initialize SDL Audio: " << SDL_GetError();
		use_sound = false;
		sys().add_console(oss.str());
		return;
	}

	int audio_rate = 44100; // 22050;
	Uint16 audio_format = AUDIO_S16SYS;
	int audio_channels = 2;
	int audio_buffers = 4096;

	if (Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers) != 0) {
		ostringstream oss;
		oss << "Unable to initialize audio: " << Mix_GetError();
		sys().add_console(oss.str());
		return;
	}
}



void sound::deinit()
{
	if (!use_sound) return;
	Mix_CloseAudio();
}
