/*
Danger from the Deep - Open source submarine simulation
Copyright (C) 2003-2006  Thorsten Jordan, Luis Barrancos and others.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// A tokenizer
// subsim (C)+(W) Thorsten Jordan. SEE LICENSE

#include <sstream>
#include <algorithm>

#include "tokenizer.h"

using namespace std;

tokenizer::tokenizer(const string& filename, vector<token>& tokenlist_,
	int token_none_, int token_string_, int token_number_, int token_id_) :
	inputstr(filename.c_str()), tokenlist(tokenlist_),
	token_none(token_none_), token_string(token_string_),
	token_number(token_number_), token_id(token_id_), line(1), column(0)
{
	sort(tokenlist.begin(), tokenlist.end());
	read_next();
}

void tokenizer::count(char c)
{
	switch(c) { 
 		case '\t': column += 8; break;
		case '\n': ++line; column = 0; break;
	        default: ++column;
	}
}

token tokenizer::read_next()
{
	previous = current;
	current = token(token_none, "");
	vector<token>::iterator pos = tokenlist.begin(), pos2;
	unsigned curr_len = 0, len = 0;
	bool can_be_id = false;
	bool comment = false;
	char c;

	while (inputstr.get(c)) {
		if (comment) {
			if (c == '\n') {
				comment = false;
			} else {
				count(c);
				continue;
			}
		}
		if (c == '/') {	// check for comments
			char c2;
			if (!inputstr.get(c2)) {
				inputstr.putback(c);
			} else {
				if (c2 == '/') {
					comment = true;
					count(c);
					count(c2);
					continue;
				} else {
					inputstr.putback(c2);
				}
			}
		}
		if (current.type == token_string) {
			if ((unsigned char)(c) < 32) {
				inputstr.putback(c);
				current.type = token_none;
				return current; // not recognized
			}
			count(c);
			current.text += c;
			if (c == '\"')
				return current;
		} else if (current.type == token_number) {
			if (!isdigit(c)) {
				inputstr.putback(c);
				return current;
			} else {
				count(c);
				current.text += c;
			}
		} else if (current.type == token_id) {
			if (isalnum(c) || c == '_') {
				count(c);
				current.text += c;
			} else {
				inputstr.putback(c);
				return current;
			}
		} else if (current.type == token_none) {
			count(c);
			if (isspace(c))
				continue;
			current.text += c;
			if (isdigit(c)) {
				current.type = token_number;
			} else if (c == '\"') {
				current.type = token_string;
			} else {
				if (isalpha(c) || c == '_')
					can_be_id = true;
				pos = lower_bound(tokenlist.begin(), tokenlist.end(), current);
				if (pos == tokenlist.end() || pos->text[0] != c) {
					if (can_be_id) {
						current.type = token_id;
					} else {
						return current; // not recognized
					}
				} else {
					current.type = pos->type;
					len = pos->text.length();
					curr_len = 1;
				}
			}
		} else {
			current.text += c;
			++curr_len;
			if (!isalnum(c) && c != '_')
				can_be_id = false;
			pos2 = lower_bound(pos, tokenlist.end(), current);
			bool part_of_token = true;
			if (pos != pos2) {
				if (pos2 == tokenlist.end() || pos2->text.size() < curr_len) {
					part_of_token = false;
				} else {
					int i;
					for (i = curr_len-1; i >= 0; --i)	// changes to the right are more likely
						if (pos2->text[i] != current.text[i])
							break;
					part_of_token = (i < 0);
				}
			} else if (curr_len > len || current.text[curr_len-1] != pos->text[curr_len-1]) {
				part_of_token = false;
			}
			if (part_of_token) {
				pos = pos2;
				len = pos2->text.size();
			} else {
				if (can_be_id) {
					current.type = token_id;
				} else if (curr_len-1 == len) {
					inputstr.putback(c);
					current = *pos;
					return current;
				} else {
					count(c);
					return current; // not recognized
				}
			}
			count(c);
		}
	}
	if (current.type == token_string) {	// string not closed
		current.type = token_none;
	} else if (current.type != token_id && current.type != token_number) {
		// token is from some type
		if (pos == tokenlist.end() || pos->text != current.text) {
			if (can_be_id)
				current.type = token_id;
			else
				current.type = token_none; // token not complete
		}
	}
	return current;
}
