/*
Danger from the Deep - Open source submarine simulation
Copyright (C) 2003-2006  Thorsten Jordan, Luis Barrancos and others.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// (ocean) water simulation and display (OpenGL)
// subsim (C)+(W) Thorsten Jordan. SEE LICENSE

// this file contains special x86-SSE optimized functions.
// nearly 40% of cpu time is consumed in ONE function,
// (compute_coord), so this is optimized with SSE.

#ifdef USE_SSE

#include "water_sse.h"

/*
A note about possible optimizations:
- implement the shorter code of writing the results from the intrinsic code
  also for hand written asm (saves only a few instructions per vertex-quadruple)
  --> only little win
- memalign the result array on 16-byte boundaries and use movaps instead of movups
  --> benefit unclear, can't be much
- use 4x16bit Uint MMX instead of 2x32bit Uint to save instructions. Conversion to
  and from 4xfloat SSE need 32bit Uint, so this won't save much. But it is done
  four times per vertex quadruple.
  --> could be worth the work
- same for general MMX usage.
  --> could spare some instructions, but not really much.
- Shorten function to fetch one height (get_height_and_displacement_sse_i) so that the
  compiler inlines it (or force inlining).
  Analyze shows that the function has even less effective assembler instructions than the
  hand written one. Use -finline-limit=1200 to force inlining.
  --> This could save a lot performance, as the function is called four times per
      vertex quadruple. It seems to be longer than the asm version because the
      compiler doesn't inline it. Its around 60-70 instructions, the hand written
      asm version around 50-60. No, it is shorter, but longer with the call/stack management
      overhead.
- The hand written asm version makes clever use of registers and uses less memory
  accesses or reads than the intrinsic version. This could be accomplished with
  the intrinsic version by caching the constants in user defined additional variables.
  to be tested... i doubt that this helps much.
*/

bool x86_sse_supported()
{
#ifdef USE_SSE_ALWAYS
	return true;
#else
	Uint32 cpuid[4];
	asm volatile ("movl $1, %%eax		\n\t"
		      "cpuid			\n\t"
		      "mov %%eax, (%%esi)	\n\t"
		      "mov %%ebx, 4(%%esi)	\n\t"
		      "mov %%ecx, 8(%%esi)	\n\t"
		      "mov %%edx, 12(%%esi)	\n\t"
		      : : "S" (&cpuid) : "eax", "ebx", "ecx", "edx" );
	// bit# 23: mmx
	// bit# 25: sse
	// bit# 26: sse2
	return ((cpuid[3] >> 25) & 1) != 0;
#endif
}



static const float fac_1_64_and_1_128[8] __attribute__((aligned(16))) = {
	1.0f/64, 1.0f/64, 1.0f/64, 1.0f/64,
	1.0f/128, 1.0f/128, 1.0f/128, 1.0f/128 };
static float v_sse[8] __attribute__((aligned(16)));
static float vadd_sse[8] __attribute__((aligned(16)));
static float vfrac_sse[8] __attribute__((aligned(16)));
static float vfracadd_sse[8] __attribute__((aligned(16)));
static float xyfrac2_sse[8] __attribute__((aligned(16)));
static const float one_sse[8] __attribute__((aligned(16))) = { 1.0f, 1.0f, 1.0f, 1.0f };
// const while water object exist:
// static float rcpwavelength[4] __attribute__((aligned(16)));
// static float waveresolution_sse[4] __attribute__((aligned(16)));
static const float vadd_fac01234444[8] __attribute__((aligned(16))) = { 0.0f, 1.0f, 2.0f, 3.0f,
									4.0f, 4.0f, 4.0f, 4.0f };
static const float fac2_16[4] __attribute__((aligned(16))) = {
	65536.0f, 65536.0f, 65536.0f, 65536.0f  };
static const Uint32 helpervalues[8] = { 0xfff, 0xfff, 0x3ff, 0x3ff, 2048, 2048, 512, 512 };
static Uint32 x0y0x1y1[16];
static Uint32* tmpptr[4];
static const Uint32 one_2x32[2] = { 0x1, 0x1 };
static Uint32 wave_res_and_2x[2];
static Uint32 waveresshiftq[2] = { 0, 0 };
static const Uint32 fac2_16i[2] = { 65536, 65536 };

// give four indices in mm0/mm1, fills xmm5/6/7 with heights/displx/disply
// be careful! it trashes mm0-7 !
static inline void get_height_and_displacement_sse(const Uint32* baseptr)
{
	// mm0/mm1 hold indices for four coordinates to load height/displx/disply from
	asm volatile ("pslld $2, %%mm0		\n\t"	// *4 to compute index of 32bit-words
		      "pslld $2, %%mm1		\n\t"	// *4 to compute index of 32bit-words
		      "movl %%eax, (%%esi)	\n\t"	// store base ptr address
		      "movl %%eax, 4(%%esi)	\n\t"	// store base ptr address, second dword
		      "movq (%%esi), %%mm2	\n\t"	// load base ptr address
		      "paddd %%mm2, %%mm0	\n\t"	// add address
		      "paddd %%mm2, %%mm1	\n\t"	// add address
		      "movq %%mm0, (%%esi)	\n\t"	// store addresses 0 and 1
		      "movq %%mm1, 8(%%esi)	\n\t"	// store addresses 2 and 3
		      : : "a" (baseptr), "S" (tmpptr)
		      : /* mm0-1 */ );

	asm volatile ("movl (%%esi), %%eax	\n\t"	// load address 0
		      "movl 4(%%esi), %%ebx	\n\t"	// load address 1
		      "movl 8(%%esi), %%ecx	\n\t"	// load address 2
		      "movl 12(%%esi), %%edx	\n\t"	// load address 3
		      "movl (%%eax), %%eax	\n\t"	// load value at address 0
		      "movl (%%ebx), %%ebx	\n\t"	// load value at address 1
		      "movl (%%ecx), %%ecx	\n\t"	// load value at address 2
		      "movl (%%edx), %%edx	\n\t"	// load value at address 3
		      "movl %%eax, (%%esi)	\n\t"	// store value of address 0
		      "movl %%ebx, 4(%%esi)	\n\t"	// store value of address 1
		      "movl %%ecx, 8(%%esi)	\n\t"	// store value of address 2
		      "movl %%edx, 12(%%esi)	\n\t"	// store value of address 3
		      "movq (%%esi), %%mm0	\n\t"	// load values of address 0 and 1
		      "movq 8(%%esi), %%mm1	\n\t"	// load values of address 2 and 3
		      "movq %%mm0, %%mm2	\n\t"	// copy for displx
		      "movq %%mm1, %%mm3	\n\t"	// copy for displx
		      "psrld $22, %%mm2		\n\t"	// >> 22 for displx
		      "psrld $22, %%mm3		\n\t"	// >> 22 for displx
		      "movq %%mm0, %%mm4	\n\t"	// copy for disply
		      "movq %%mm1, %%mm5	\n\t"	// copy for disply
		      "psrld $12, %%mm4		\n\t"	// >> 12 for disply
		      "psrld $12, %%mm5		\n\t"	// >> 12 for disply
		      "movq (%%edi), %%mm6	\n\t"	// load 0xfff
		      "movq 8(%%edi), %%mm7	\n\t"	// load 0x3ff
		      "pand %%mm6, %%mm0	\n\t"	// and with 0xfff
		      "pand %%mm6, %%mm1	\n\t"	// and with 0xfff
		      "pand %%mm7, %%mm2	\n\t"	// and with 0x3ff
		      "pand %%mm7, %%mm3	\n\t"	// and with 0x3ff
		      "pand %%mm7, %%mm4	\n\t"	// and with 0x3ff
		      "pand %%mm7, %%mm5	\n\t"	// and with 0x3ff
		      "movq 16(%%edi), %%mm6	\n\t"	// load 2048
		      "movq 24(%%edi), %%mm7	\n\t"	// load 512
		      "psubd %%mm6, %%mm0	\n\t"	// sub 2048
		      "psubd %%mm6, %%mm1	\n\t"	// sub 2048
		      "psubd %%mm7, %%mm2	\n\t"	// sub 512
		      "psubd %%mm7, %%mm3	\n\t"	// sub 512
		      "psubd %%mm7, %%mm4	\n\t"	// sub 512
		      "psubd %%mm7, %%mm5	\n\t"	// sub 512
		      // now mm0-5 hold values for height/displx/disply
		      "cvtpi2ps %%mm1, %%xmm5	\n\t"	// to float
		      "cvtpi2ps %%mm3, %%xmm6	\n\t"	// to float
		      "cvtpi2ps %%mm5, %%xmm7	\n\t"	// to float
		      "movlhps %%xmm5, %%xmm5	\n\t"	// and in upper 2 floats
		      "movlhps %%xmm6, %%xmm6	\n\t"	// and in upper 2 floats
		      "movlhps %%xmm7, %%xmm7	\n\t"	// and in upper 2 floats
		      "cvtpi2ps %%mm0, %%xmm5	\n\t"	// to float
		      "cvtpi2ps %%mm2, %%xmm6	\n\t"	// to float
		      "cvtpi2ps %%mm4, %%xmm7	\n\t"	// to float
		      : : "S" (tmpptr), "D" (helpervalues)
		      : "eax", "ebx", "ecx", "edx" /* and mm0-7*/ );

	// now multiply with factors
	asm volatile ("mulps 16(%%eax), %%xmm5	\n\t"	// multiply with 16/2048=1/128
		      "mulps (%%eax), %%xmm6	\n\t"	// multiply with 8/512=1/64
		      "mulps (%%eax), %%xmm7	\n\t"	// multiply with 8/512=1/64
		      : : "a" (fac_1_64_and_1_128) : );
}



void water_compute_coord_init_sse(float /*wavetile_length_rcp*/, unsigned wave_resolution,
				  unsigned wave_resolution_shift)
{
// 	rcpwavelength[0] = rcpwavelength[1] = rcpwavelength[2] = rcpwavelength[3] =
// 		wavetile_length_rcp;
// 	waveresolution_sse[0] = waveresolution_sse[1] = waveresolution_sse[2] = waveresolution_sse[3] = 
// 		float(wave_resolution);
	wave_res_and_2x[0] = wave_res_and_2x[1] = wave_resolution-1;
	waveresshiftq[0] = wave_resolution_shift;
}



void water_compute_coord_1line_sse(const vector2f& v, const vector2f& vadd,
				   const vector2f& vfrac, const vector2f& vfracadd,
				   const Uint32* tiledata_baseptr, float minusviewposz,
				   float* coordptr, unsigned xres)
{
	// preparation part 1: fill v.x/y, vadd.x/y
	asm volatile ("movss (%%eax), %%xmm0		\n\t"	// load v.x
		      "movss 4(%%eax), %%xmm1		\n\t"	// load v.y
		      "shufps $0x00, %%xmm0, %%xmm0	\n\t"	// set v.x to all 4 floats
		      "shufps $0x00, %%xmm1, %%xmm1	\n\t"	// set v.y to all 4 floats
		      "movss (%%ebx), %%xmm2		\n\t"	// load vadd.x
		      "movss 4(%%ebx), %%xmm3		\n\t"	// load vadd.y
		      "shufps $0x00, %%xmm2, %%xmm2	\n\t"	// set vadd.x to all 4 floats
		      "shufps $0x00, %%xmm3, %%xmm3	\n\t"	// set vadd.y to all 4 floats
		      "movaps 16(%%edx), %%xmm4		\n\t"	// load 4
		      "movaps %%xmm4, %%xmm5		\n\t"	// load 4
		      "mulps %%xmm2, %%xmm4		\n\t"	// vadd.x*4
		      "mulps %%xmm3, %%xmm5		\n\t"	// vadd.x*4
		      "movaps %%xmm4, (%%ecx)		\n\t"	// store vadd.x*4 ps
		      "movaps %%xmm5, 16(%%ecx)		\n\t"	// store vadd.y*4 ps
		      "mulps (%%edx), %%xmm2		\n\t"	// mul vadd.x with 0/1/2/3
		      "mulps (%%edx), %%xmm3		\n\t"	// mul vadd.y with 0/1/2/3
		      "addps %%xmm2, %%xmm0		\n\t"	// add to v.x to get start vector
		      "addps %%xmm3, %%xmm1		\n\t"	// add to v.y to get start vector
		      "movaps %%xmm0, (%%esi)		\n\t"	// store v_sse
		      "movaps %%xmm1, 16(%%esi)		\n\t"	// store v_sse
		      : /* output */
		      : /* input */ "a" (&v.x), "b" (&vadd.x), "c" (vadd_sse),
			"d" (vadd_fac01234444), "S" (v_sse)
		      : /* manipulated */ );

	// preparation part 2: fill vfrac.x/y, vfracadd.x/y
	asm volatile ("movss (%%eax), %%xmm0		\n\t"	// load vfrac.x
		      "movss 4(%%eax), %%xmm1		\n\t"	// load vfrac.y
		      "shufps $0x00, %%xmm0, %%xmm0	\n\t"	// set vfrac.x to all 4 floats
		      "shufps $0x00, %%xmm1, %%xmm1	\n\t"	// set vfrac.y to all 4 floats
		      "movss (%%ebx), %%xmm2		\n\t"	// load vfracadd.x
		      "movss 4(%%ebx), %%xmm3		\n\t"	// load vfracadd.y
		      "shufps $0x00, %%xmm2, %%xmm2	\n\t"	// set vfracadd.x to all 4 floats
		      "shufps $0x00, %%xmm3, %%xmm3	\n\t"	// set vfracadd.y to all 4 floats
		      "movaps 16(%%edx), %%xmm4		\n\t"	// load 4
		      "movaps %%xmm4, %%xmm5		\n\t"	// load 4
		      "mulps %%xmm2, %%xmm4		\n\t"	// vfracadd.x*4
		      "mulps %%xmm3, %%xmm5		\n\t"	// vfracadd.x*4
		      "movaps %%xmm4, (%%ecx)		\n\t"	// store vfracadd.x*4 ps
		      "movaps %%xmm5, 16(%%ecx)		\n\t"	// store vfracadd.y*4 ps
		      "mulps (%%edx), %%xmm2		\n\t"	// mul vfracadd.x with 0/1/2/3
		      "mulps (%%edx), %%xmm3		\n\t"	// mul vfracadd.y with 0/1/2/3
		      "addps %%xmm2, %%xmm0		\n\t"	// add to vfrac.x to get start vector
		      "addps %%xmm3, %%xmm1		\n\t"	// add to vfrac.y to get start vector
		      "movaps %%xmm0, (%%esi)		\n\t"	// store vfrac_sse
		      "movaps %%xmm1, 16(%%esi)		\n\t"	// store vfrac_sse
		      : /* output */
		      : /* input */ "a" (&vfrac.x), "b" (&vfracadd.x), "c" (vfracadd_sse),
			"d" (vadd_fac01234444), "S" (vfrac_sse)
		      : /* manipulated */ );

/*
  sometimes SSE and C++ generated values differ by a large amount (~ 0.5... 1 total!)
  SSE is much less precise because all operands are stored with only float precision
  leading to serious rounding errors. But can it be THAT unprecise?
  yes, maybe: small rounding error in x/y direction can lead to large jumps in per tile x/y values
  for far coordinates. With larger variance of x/y coordinates different height values can be fetched for
  far coordinates, leading to greater differences in height values.
*/

	// note! works only if xres+1 is multiple of 4! that means xres % 4 = 3
	unsigned xloopc = (xres+1)/4;
	for (unsigned xx = 0; xx < xloopc; ++xx, coordptr += 12) {

		// compute xfrac2/yfrac2 to xmm0/xmm1
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load vfrac.x
			      "movaps 16(%%eax), %%xmm1		\n\t"	// load vfrac.y
			      "cvttps2pi %%xmm0, %%mm4		\n\t"	// trunc to integer in mm4
			      "cvttps2pi %%xmm1, %%mm6		\n\t"	// trunc to integer in mm6
			      "movhlps %%xmm0, %%xmm2		\n\t"	// move 2 upper floats to 2 lower in xmm2
			      "movhlps %%xmm1, %%xmm3		\n\t"	// move 2 upper floats to 2 lower in xmm3
			      "cvttps2pi %%xmm2, %%mm5		\n\t"	// trunc to integer in mm5
			      "cvttps2pi %%xmm3, %%mm7		\n\t"	// trunc to integer in mm7
			      "movq (%%ecx), %%mm0		\n\t"	// load wave_resolution-1
			      "movq %%mm0, %%mm1		\n\t"	// load wave_resolution-1
			      "movq %%mm0, %%mm2		\n\t"	// load wave_resolution-1
			      "movq %%mm0, %%mm3		\n\t"	// load wave_resolution-1
			      "pand %%mm4, %%mm0		\n\t"	// mask out tile coordinates
			      "pand %%mm5, %%mm1		\n\t"	// mask out tile coordinates
			      "pand %%mm6, %%mm2		\n\t"	// mask out tile coordinates
			      "pand %%mm7, %%mm3		\n\t"	// mask out tile coordinates
			      "psubd %%mm0, %%mm4		\n\t"	// subtract masked out part
			      "psubd %%mm1, %%mm5		\n\t"	// " | mm4-7 will then hold the part
			      "psubd %%mm2, %%mm6		\n\t"	// " | of coordinates that are greater
			      "psubd %%mm3, %%mm7		\n\t"	// " | than one tile (when vfrac+vfracadd > 1tile)
			      "cvtpi2ps %%mm5, %%xmm2		\n\t"	// back to float
			      "cvtpi2ps %%mm7, %%xmm3		\n\t"	// back to float
			      "movlhps %%xmm2, %%xmm2		\n\t"	// and in upper 2 floats
			      "movlhps %%xmm3, %%xmm3		\n\t"	// and in upper 2 floats
			      "cvtpi2ps %%mm4, %%xmm2		\n\t"	// back to float
			      "cvtpi2ps %%mm6, %%xmm3		\n\t"	// back to float
			      "subps %%xmm2, %%xmm0		\n\t"	// subtract >1tile part from vfrac
			      "subps %%xmm3, %%xmm1		\n\t"	// subtract >1tile part from vfrac
			      "movaps %%xmm0, (%%eax)		\n\t"	// store corrected vfrac.x
			      "movaps %%xmm1, 16(%%eax)		\n\t"	// store corrected vfrac.y
			      "cvtpi2ps %%mm1, %%xmm2		\n\t"	// back to float for x0/y0
			      "cvtpi2ps %%mm3, %%xmm3		\n\t"	// back to float
			      "movlhps %%xmm2, %%xmm2		\n\t"	// and in upper 2 floats
			      "movlhps %%xmm3, %%xmm3		\n\t"	// and in upper 2 floats
			      "cvtpi2ps %%mm0, %%xmm2		\n\t"	// back to float
			      "cvtpi2ps %%mm2, %%xmm3		\n\t"	// back to float
			      "subps %%xmm2, %%xmm0		\n\t"	// now xmm0 holds its fractional part
			      "subps %%xmm3, %%xmm1		\n\t"	// now xmm1 holds its fractional part
			      // now mm0/1 holds x0, mm2/3 holds y0

			      // compute x1/y1 in mm4-7
			      "movq (%%ebx), %%mm4		\n\t"	// load 1
			      "movq %%mm4, %%mm5		\n\t"	// load 1
			      "movq %%mm4, %%mm6		\n\t"	// load 1
			      "movq %%mm4, %%mm7		\n\t"	// load 1
			      "paddd %%mm0, %%mm4		\n\t"	// x1=1+x0
			      "paddd %%mm1, %%mm5		\n\t"	// x1=1+x0
			      "paddd %%mm2, %%mm6		\n\t"	// y1=1+y0
			      "paddd %%mm3, %%mm7		\n\t"	// y1=1+y0
			      "pand (%%ecx), %%mm4 		\n\t"	// and width-1
			      "pand (%%ecx), %%mm5 		\n\t"	// and width-1
			      "pand (%%ecx), %%mm6 		\n\t"	// and width-1
			      "pand (%%ecx), %%mm7 		\n\t"	// and width-1

			      // shift y0/y1 by wave_resolution_shift
			      "pslld (%%edx), %%mm2		\n\t"	// shift y0
			      "pslld (%%edx), %%mm3		\n\t"	// shift y0
			      "pslld (%%edx), %%mm6		\n\t"	// shift y1
			      "pslld (%%edx), %%mm7		\n\t"	// shift y1

			      // save x0/y0/x1/y1
			      "movq %%mm0, (%%esi)		\n\t"	// save x0
			      "movq %%mm1, 8(%%esi)		\n\t"	// save x0
			      "movq %%mm2, 16(%%esi)		\n\t"	// save y0
			      "movq %%mm3, 24(%%esi)		\n\t"	// save y0
			      "movq %%mm4, 32(%%esi)		\n\t"	// save x1
			      "movq %%mm5, 40(%%esi)		\n\t"	// save x1
			      "movq %%mm6, 48(%%esi)		\n\t"	// save y1
			      "movq %%mm7, 56(%%esi)		\n\t"	// save y1
			      : : "a" (vfrac_sse), "b" (one_2x32), "c" (wave_res_and_2x),
				"d" (waveresshiftq), "S" (x0y0x1y1) : );

		// save xfrac2/yfrac2
		asm volatile ("movaps %%xmm0, (%%eax)		\n\t" // store xfrac2
			      "movaps %%xmm1, 16(%%eax)		\n\t" // store yfrac2
			      "movaps (%%ebx), %%xmm0		\n\t" // load 1.0
			      "movaps (%%ebx), %%xmm1		\n\t" // load 1.0
			      "subps (%%eax), %%xmm0		\n\t" // 1.0-xfrac2
			      "subps 16(%%eax), %%xmm1		\n\t" // 1.0-yfrac2
			      : : "a" (xyfrac2_sse), "b" (one_sse) : );

		// compute i0 etc. xmm0/1 holds xfrac2/yfrac2
		asm volatile ("paddd %%mm2, %%mm0		\n\t"	// x0+y0
			      "paddd %%mm3, %%mm1		\n\t"	// x0+y0
			      : : : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("mulps %%xmm1, %%xmm0		\n\t"	// comp (1-xfrac2)*(1-yfrac2)
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "movaps %%xmm5, %%xmm2		\n\t"
			      "movaps %%xmm6, %%xmm3		\n\t"
			      "movaps %%xmm7, %%xmm4		\n\t"
			      : : : );

		// compute i1 etc.
		asm volatile ("movq 32(%%eax), %%mm0		\n\t"	// get x1
			      "movq 40(%%eax), %%mm1		\n\t"	// get x1
			      "paddd 16(%%eax), %%mm0		\n\t"	// add y0
			      "paddd 24(%%eax), %%mm1		\n\t"	// add y0
			      : : "a" (x0y0x1y1) : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load xfrac2
			      "mulps %%xmm1, %%xmm0		\n\t"	// comp xfrac2*(1-yfrac2)
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "addps %%xmm5, %%xmm2		\n\t"
			      "addps %%xmm6, %%xmm3		\n\t"
			      "addps %%xmm7, %%xmm4		\n\t"
			      : : "a" (xyfrac2_sse) : );

		// compute i2 etc.
		asm volatile ("movq (%%eax), %%mm0			\n\t"	// get x0
			      "movq 8(%%eax), %%mm1			\n\t"	// get x0
			      "paddd 48(%%eax), %%mm0		\n\t"	// add y1
			      "paddd 56(%%eax), %%mm1		\n\t"	// add y1
			      : : "a" (x0y0x1y1) : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("movaps (%%ebx), %%xmm0		\n\t"	// load 1
			      "movaps 16(%%eax), %%xmm1		\n\t"	// load yfrac2
			      "subps (%%eax), %%xmm0		\n\t"	// comp 1-xfrac2
			      "mulps %%xmm1, %%xmm0		\n\t"	// comp (1-xfrac2)*yfrac
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "addps %%xmm5, %%xmm2		\n\t"
			      "addps %%xmm6, %%xmm3		\n\t"
			      "addps %%xmm7, %%xmm4		\n\t"
			      : : "a" (xyfrac2_sse), "b" (one_sse) : );

		// compute i3 etc.
		asm volatile ("movq 32(%%eax), %%mm0		\n\t"	// get x1
			      "movq 40(%%eax), %%mm1		\n\t"	// get x1
			      "paddd 48(%%eax), %%mm0		\n\t"	// add y1
			      "paddd 56(%%eax), %%mm1		\n\t"	// add y1
			      : : "a" (x0y0x1y1) : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load xfrac2
			      "mulps %%xmm1, %%xmm0		\n\t"	// comp xfrac2*yfrac2
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "addps %%xmm5, %%xmm2		\n\t"
			      "addps %%xmm6, %%xmm3		\n\t"
			      "addps %%xmm7, %%xmm4		\n\t"
			      : : "a" (xyfrac2_sse) : );

		// now add xyzpos, that means add vx/vy to xmm3/4 and -viewpos.z to xmm2
		asm volatile ("movss (%%eax), %%xmm5		\n\t"	// load -viewpos.z
			      "shufps $0x00, %%xmm5, %%xmm5	\n\t"	// load to all ps
 			      "addps (%%ebx), %%xmm3		\n\t"	// add v.x
 			      "addps 16(%%ebx), %%xmm4		\n\t"	// add v.y
			      "addps %%xmm5, %%xmm2		\n\t"	// add -viewpos.z
			      : : "a" (&minusviewposz), "b" (v_sse) : );

		// now xmm2-4 holds interpolated height/displx/disply
		// coordptr points to first of 3*4 floats to store results to.
		// xmm2-4 hold values    : z3 z2 z1 z0 | x3 x2 x1 x0 | y3 y2 y1 y0
		// stored should be      : x1 z0 y0 x0 | y2 x2 z1 y1 | z3 y3 x3 z2
		// intermediate form     : x1 x0 z0 y0 | y2 y1 z1 x2 | y3 x3 z3 z2
		// mix registers in xmm5-7 and store result
		asm volatile ("movaps %%xmm4, %%xmm5		\n\t" //5: y3y2y1y0, copy
			      "movaps %%xmm2, %%xmm6		\n\t" //6: z3z2z1z0, copy
			      "unpcklps %%xmm2, %%xmm5		\n\t" //5: z1y1z0y0, mix z0y0
			      "shufps $0xe4, %%xmm3, %%xmm6	\n\t" //6: x3x2z1z0, mix z1x2 part1
			      "movlhps %%xmm3, %%xmm5		\n\t" //5: x1x0z0y0, copy x1x0
			      "shufps $0x66, %%xmm6, %%xmm6	\n\t" //6: z1x2z1x2, mix z1x2 part2
			      "movaps %%xmm3, %%xmm7		\n\t" //7: x3x2x1x0, copy
			      "shufps $0xd2, %%xmm5, %%xmm5	\n\t" //5: x1z0y0x0, xmm5 ready
			      "unpckhps %%xmm4, %%xmm7		\n\t" //7: y3x3y2x2, get y3x3
			      "shufps $0x94, %%xmm4, %%xmm6	\n\t" //6: y2y1z1x2, copy y2y1
			      "movhlps %%xmm2, %%xmm7		\n\t" //7: y3x3z3z2, copy z3z2
			      "shufps $0xc6, %%xmm6, %%xmm6	\n\t" //6: y2x2z1y1, xmm6 ready
			      "shufps $0x78, %%xmm7, %%xmm7	\n\t" //7: z3y3x3z2, xmm7 ready
			      "movups %%xmm5, (%%eax)		\n\t" // store bytes 0-3
			      "movups %%xmm6, 16(%%eax)		\n\t" // store bytes 4-7
			      "movups %%xmm7, 32(%%eax)		\n\t" // store bytes 8-11
			      : : "a" (coordptr) : );

		// increase v/vfrac
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load vadd.x
			      "movaps 16(%%eax), %%xmm1		\n\t"	// load vadd.y
			      "movaps (%%ecx), %%xmm2		\n\t"	// load vfracadd.x
			      "movaps 16(%%ecx), %%xmm3		\n\t"	// load vfracadd.y
			      "addps (%%ebx), %%xmm0		\n\t"	// add to v.x
			      "addps 16(%%ebx), %%xmm1		\n\t"	// add to v.y
			      "addps (%%edx), %%xmm2		\n\t"	// add to vfrac.x
			      "addps 16(%%edx), %%xmm3		\n\t"	// add to vfrac.y
			      "movaps %%xmm0, (%%ebx)		\n\t"	// store new v.x
			      "movaps %%xmm1, 16(%%ebx)		\n\t"	// store new v.y
			      "movaps %%xmm2, (%%edx)		\n\t"	// store new vfrac.x
			      "movaps %%xmm3, 16(%%edx)		\n\t"	// store new vfrac.y
			      : /* output */
			      : /* input */ "a" (vadd_sse), "b" (v_sse), "c" (vfracadd_sse), "d" (vfrac_sse)
			      : /* manipulated */ );
	}

	asm volatile ("emms		\n\t");
}

#ifdef USE_SSE_INTRINSICS

#include <xmmintrin.h>

union packed_Uint32
{
	Uint32 val_uint32[2];
	__m64 val_packed;
};

// note: the _mm_set_* functions store arguments in MSB order, opposite to _mm_loadu_ps etc.
static const __m64 val0x3ff = _mm_set1_pi32(0x03ff);
static const __m64 val0xfff = _mm_set1_pi32(0x0fff);
static const __m64 val2048 = _mm_set1_pi32(2048);
static const __m64 val512 = _mm_set1_pi32(512);
static const __m128 val1by64 = _mm_set1_ps(1.0f/64);
static const __m128 val1by128 = _mm_set1_ps(1.0f/128);
static const __m128 val4_0 = _mm_set1_ps(4.0f);
static const __m128 val1_0 = _mm_set1_ps(1.0f);
static const __m128 zero_one_two_three = _mm_setr_ps(0.0f, 1.0f, 2.0f, 3.0f);
static const __m64 val1 = _mm_set1_pi32(1);
static __m64 wave_res_and_2xpi32;
static __m64 waveresshiftq_m64;

// convert packed integers [A1 A0] [B1 B0] to [B1 B0 A1 A0]
static inline __m128 convert_pi32_to_m128(__m64 low, __m64 high)
{
	__m128 r;
	r = _mm_cvtpi32_ps(r, high);
	r = _mm_movelh_ps(r, r);
	r = _mm_cvtpi32_ps(r, low);
	return r;
}

// convert and truncate packed floats [A3 A2 A1 A0] to [A1 A0] and [A3 A2]
static inline void convert_trunc_m128_to_pi32(__m128 a, __m64& low, __m64& high)
{
	low = _mm_cvttps_pi32(a);
	a = _mm_movehl_ps(a, a);
	high = _mm_cvttps_pi32(a);
}


void water_compute_coord_init_sse_i(float /*wavetile_length_rcp*/, unsigned wave_resolution,
				  unsigned wave_resolution_shift)
{
// 	rcpwavelength = _mm_set1_ps(wavetile_length_rcp);
// 	waveresolution_sse = _mm_set1_ps(float(wave_resolution));
	wave_res_and_2xpi32 = _mm_set1_pi32(wave_resolution-1);
	waveresshiftq_m64 = _mm_cvtsi32_si64(wave_resolution_shift);
}



static inline void get_height_and_displacement_sse_i(const Uint32* baseptr, __m64 coord01, __m64 coord23,
						     __m128& x, __m128& y, __m128& z)
{
	packed_Uint32 coord01_tmp, coord23_tmp;
	coord01_tmp.val_packed = coord01;
	coord23_tmp.val_packed = coord23;
	packed_Uint32 value_tmp01, value_tmp23;
	value_tmp01.val_uint32[0] = baseptr[coord01_tmp.val_uint32[0]];
	value_tmp01.val_uint32[1] = baseptr[coord01_tmp.val_uint32[1]];
	value_tmp23.val_uint32[0] = baseptr[coord23_tmp.val_uint32[0]];
	value_tmp23.val_uint32[1] = baseptr[coord23_tmp.val_uint32[1]];
	// fixme: we don't need to mask out bits for displx here, 32bit>>22 are already 10 bits
	// working with 4x16bit in mmx registers doesn't spare much. The conversions
	// between 2x32 and 4x16 would eat the same number of instructions as the current
	// computations.
	__m64 displx01 = _mm_sub_pi32(/*_mm_and_si64(*/_mm_srli_pi32(value_tmp01.val_packed, 22)/*, val0x3ff)*/, val512);
	__m64 displx23 = _mm_sub_pi32(/*_mm_and_si64(*/_mm_srli_pi32(value_tmp23.val_packed, 22)/*, val0x3ff)*/, val512);
	__m64 disply01 = _mm_sub_pi32(_mm_and_si64(_mm_srli_pi32(value_tmp01.val_packed, 12), val0x3ff), val512);
	__m64 disply23 = _mm_sub_pi32(_mm_and_si64(_mm_srli_pi32(value_tmp23.val_packed, 12), val0x3ff), val512);
	__m64 height01 = _mm_sub_pi32(_mm_and_si64(value_tmp01.val_packed, val0xfff), val2048);
	__m64 height23 = _mm_sub_pi32(_mm_and_si64(value_tmp23.val_packed, val0xfff), val2048);

	x = _mm_mul_ps(convert_pi32_to_m128(displx01, displx23), val1by64);
	y = _mm_mul_ps(convert_pi32_to_m128(disply01, disply23), val1by64);
	z = _mm_mul_ps(convert_pi32_to_m128(height01, height23), val1by128);
}

void water_compute_coord_1line_sse_i(const vector2f& v, const vector2f& vadd,
				     const vector2f& vfrac, const vector2f& vfracadd,
				     const Uint32* tiledata_baseptr, float minusviewposz,
				     float* coordptr, unsigned xres)
{
	// load v_x/y, vadd_x/y
	__m128 v_x, v_y, vadd_x, vadd_y;
	vadd_x = _mm_load1_ps(&vadd.x);
	vadd_y = _mm_load1_ps(&vadd.y);
	v_x = _mm_add_ps(_mm_load1_ps(&v.x), _mm_mul_ps(vadd_x, zero_one_two_three));
	v_y = _mm_add_ps(_mm_load1_ps(&v.y), _mm_mul_ps(vadd_y, zero_one_two_three));
	vadd_x = _mm_mul_ps(vadd_x, val4_0);
	vadd_y = _mm_mul_ps(vadd_y, val4_0);

	// load vfrac_x/y, vfracadd_x/y
	__m128 vfrac_x, vfrac_y, vfracadd_x, vfracadd_y;
	vfracadd_x = _mm_load1_ps(&vfracadd.x);
	vfracadd_y = _mm_load1_ps(&vfracadd.y);
	vfrac_x = _mm_add_ps(_mm_load1_ps(&vfrac.x), _mm_mul_ps(vfracadd_x, zero_one_two_three));
	vfrac_y = _mm_add_ps(_mm_load1_ps(&vfrac.y), _mm_mul_ps(vfracadd_y, zero_one_two_three));
	vfracadd_x = _mm_mul_ps(vfracadd_x, val4_0);
	vfracadd_y = _mm_mul_ps(vfracadd_y, val4_0);

	unsigned xloopc = (xres+1)/4;
	for (unsigned xx = 0; xx < xloopc; ++xx, coordptr += 12) {
		// subtract part of coordinates from vfrac_x/y that are larger than
		// one tile. After that vfrac_x/y is less than wave_resolution.
		__m64 x1_lo, x1_hi, y1_lo, y1_hi;
		// fixme: operate with 4x 16bit ints here to save registers?
		// but for conversion to sse we need 32bit registers. So it wouldn't help much...
		convert_trunc_m128_to_pi32(vfrac_x, x1_lo, x1_hi);
		convert_trunc_m128_to_pi32(vfrac_y, y1_lo, y1_hi);
		__m64 x0_lo, x0_hi, y0_lo, y0_hi;
		x0_lo = _mm_and_si64(x1_lo, wave_res_and_2xpi32);	// x0 = x1 and waveres-1
		x0_hi = _mm_and_si64(x1_hi, wave_res_and_2xpi32);
		y0_lo = _mm_and_si64(y1_lo, wave_res_and_2xpi32);
		y0_hi = _mm_and_si64(y1_hi, wave_res_and_2xpi32);
		x1_lo = _mm_sub_pi32(x1_lo, x0_lo);			// x1 = x1 - x0
		x1_hi = _mm_sub_pi32(x1_hi, x0_hi);
		y1_lo = _mm_sub_pi32(y1_lo, y0_lo);
		y1_hi = _mm_sub_pi32(y1_hi, y0_hi);
		vfrac_x = _mm_sub_ps(vfrac_x, convert_pi32_to_m128(x1_lo, x1_hi));
		vfrac_y = _mm_sub_ps(vfrac_y, convert_pi32_to_m128(y1_lo, y1_hi));
		// compute fractional part of vfrac_x/y by subtracting x0/y0
		__m128 xfrac2, yfrac2;
		xfrac2 = _mm_sub_ps(vfrac_x, convert_pi32_to_m128(x0_lo, x0_hi));
		yfrac2 = _mm_sub_ps(vfrac_y, convert_pi32_to_m128(y0_lo, y0_hi));

		// compute x1/y1, where y1 contains already shifted values,
		// so that x? + y? gives the correct index to wave height array.
		x1_lo = _mm_and_si64(_mm_add_pi32(x0_lo, val1), wave_res_and_2xpi32);
		x1_hi = _mm_and_si64(_mm_add_pi32(x0_hi, val1), wave_res_and_2xpi32);
		y1_lo = _mm_sll_pi32(_mm_and_si64(_mm_add_pi32(y0_lo, val1), wave_res_and_2xpi32), waveresshiftq_m64);
		y1_hi = _mm_sll_pi32(_mm_and_si64(_mm_add_pi32(y0_hi, val1), wave_res_and_2xpi32), waveresshiftq_m64);
		// shift left y0 so that x? + y? gives the correct index.
		y0_lo = _mm_sll_pi32(y0_lo, waveresshiftq_m64);
		y0_hi = _mm_sll_pi32(y0_hi, waveresshiftq_m64);

		// compute 1.0 - xfrac2, 1.0 - yfrac2
		__m128 xfrac2_neg = _mm_sub_ps(val1_0, xfrac2);
		__m128 yfrac2_neg = _mm_sub_ps(val1_0, yfrac2);

		// compute i0, fetch value and multiply with (1-xfrac2)*(1-yfrac2)
		__m128 x, y, z;
		get_height_and_displacement_sse_i(tiledata_baseptr, 
						_mm_add_pi32(x0_lo, y0_lo),
						_mm_add_pi32(x0_hi, y0_hi),
						x, y, z);
		__m128 factor = _mm_mul_ps(xfrac2_neg, yfrac2_neg);
		__m128 x_sum, y_sum, z_sum;
		x_sum = _mm_mul_ps(x, factor);
		y_sum = _mm_mul_ps(y, factor);
		z_sum = _mm_mul_ps(z, factor);

		// compute i1, fetch value and multiply with xfrac2*(1-yfrac2)
		get_height_and_displacement_sse_i(tiledata_baseptr, 
						_mm_add_pi32(x1_lo, y0_lo),
						_mm_add_pi32(x1_hi, y0_hi),
						x, y, z);
		factor = _mm_mul_ps(xfrac2, yfrac2_neg);
		x_sum = _mm_add_ps(x_sum, _mm_mul_ps(x, factor));
		y_sum = _mm_add_ps(y_sum, _mm_mul_ps(y, factor));
		z_sum = _mm_add_ps(z_sum, _mm_mul_ps(z, factor));

		// compute i2, fetch value and multiply with (1-xfrac2)*yfrac2
		get_height_and_displacement_sse_i(tiledata_baseptr, 
						_mm_add_pi32(x0_lo, y1_lo),
						_mm_add_pi32(x0_hi, y1_hi),
						x, y, z);
		factor = _mm_mul_ps(xfrac2_neg, yfrac2);
		x_sum = _mm_add_ps(x_sum, _mm_mul_ps(x, factor));
		y_sum = _mm_add_ps(y_sum, _mm_mul_ps(y, factor));
		z_sum = _mm_add_ps(z_sum, _mm_mul_ps(z, factor));

		// compute i3, fetch value and multiply with xfrac2*yfrac2
		get_height_and_displacement_sse_i(tiledata_baseptr, 
						_mm_add_pi32(x1_lo, y1_lo),
						_mm_add_pi32(x1_hi, y1_hi),
						x, y, z);
		factor = _mm_mul_ps(xfrac2, yfrac2);
		x_sum = _mm_add_ps(x_sum, _mm_mul_ps(x, factor));
		y_sum = _mm_add_ps(y_sum, _mm_mul_ps(y, factor));
		z_sum = _mm_add_ps(z_sum, _mm_mul_ps(z, factor));

		// now add xyzpos, that means add vx/vy to x/y_sum and -viewpos.z to z_sum
		x_sum = _mm_add_ps(x_sum, v_x);
		y_sum = _mm_add_ps(y_sum, v_y);
		z_sum = _mm_add_ps(z_sum, _mm_set1_ps(minusviewposz));

		// re-pack values in x_sum/y_sum/z_sum to 12 floats forming x/y/z triples.
		// note! we could use movaps instead of movups if coordptr is 16-byte-aligned.
		// this could be easily accomplished.
		// first
		__m128 tmp = _mm_movelh_ps(_mm_unpacklo_ps(y_sum, z_sum), x_sum); // now tmp is LSBF y0 z0 x0 x1
		_mm_storeu_ps(coordptr, _mm_shuffle_ps(tmp, tmp, _MM_SHUFFLE(3, 1, 0, 2)));
		// second
		tmp = _mm_shuffle_ps(x_sum, y_sum, _MM_SHUFFLE(2, 1, 2, 0));	// get y1,y2,x2
		__m128 tmp2 = _mm_shuffle_ps(z_sum, z_sum, _MM_SHUFFLE(1, 1, 1, 1)); // move z1 to lowest float
		tmp = _mm_move_ss(tmp, tmp2);	// now tmp is LSBF z1 x2 y1 y2
		_mm_storeu_ps(coordptr+4, _mm_shuffle_ps(tmp, tmp, _MM_SHUFFLE(3, 1, 0, 2)));
		// third
		tmp = _mm_movehl_ps(_mm_unpackhi_ps(x_sum, y_sum), z_sum); // now r2 is LSBF z2 z3 x3 y3
		_mm_storeu_ps(coordptr+8, _mm_shuffle_ps(tmp, tmp, _MM_SHUFFLE(1, 3, 2, 0)));

		// increase v_x/y and vfrac_x/y
		v_x = _mm_add_ps(v_x, vadd_x);
		v_y = _mm_add_ps(v_y, vadd_y);
		vfrac_x = _mm_add_ps(vfrac_x, vfracadd_x);
		vfrac_y = _mm_add_ps(vfrac_y, vfracadd_y);
	}

	_mm_empty();	// emms
}

#endif // USE_SSE_INTRINSICS

#endif // USE_SSE
